!> \~japanese
!! \brief 正二十面体格子上の物理場データを管理する構造型およびそれを操作・解析するための手続きを提供するモジュール.
!!
!! \par Field_IcGrid2D_Manager モジュールについて
!!   - 本モジュールでは, 正二十面格子上に水平 2 時限的に分布する物理場データを管理する構造型 Field_IcGrid2D を提供している.
!!   - 物理場データを保持する配列のメモリ確保は, initialize_Field_IcGrid2D によって構造型 Field_IcGrid2D の変数が初期化される際に行われる.
!!     また, その配列のメモリ解放は finalize_Field_IcGrid2D によって行われる.
!!   - 構造型 Field_IcGrid2D の成分である構造型 IcGrid2D_FVM のポインタ icgrid は, この物理場と関連付けられる正二十面格子データを指している.
!!   - 本モジュールでは, 正二十面格子上の物理場データを操作あるいは解析するためのシンプルな手続き(全球平均など)を提供している.
!!
!! \~english
!! \brief This module provides the derived type which manages the data of a physical field on a icosahedral grid and
!! some subroutines to manupulate and analize the field data.
!!
!! \par About Field_IcGrid2D_Manager module
!!   - 本モジュールでは, 正二十面格子上に水平 2 時限的に分布する物理場データを管理する構造型 Field_IcGrid2D を提供している.
!!   - 物理場データを保持する配列のメモリ確保は, initialize_Field_IcGrid2D によって構造型 Field_IcGrid2D の変数が初期化される際に行われる.
!!     また, その配列のメモリ解放は finalize_Field_IcGrid2D によって行われる.
!!   - 構造型 Field_IcGrid2D の成分である構造型 IcGrid2D_FVM のポインタ icgrid は, この物理場と関連付けられる正二十面格子データを指している.
!!   - 本モジュールでは, 正二十面格子上の物理場データを操作あるいは解析するためのシンプルな手続き(全球平均など)を提供している.
!!
!! \~
!!
!! <br><br>
!! \b Copyright (C) GFD Dennou Club, 2011-2012. All rights reserved. <br>
!! \b license ?? <br>
!! \author Yuta Kawai
!!
module Field_IcGrid2D_Manager

  ! モジュール引用; Use statement
  !

  ! 種類型パラメータ
  ! Kind type parameter
  !
  use dc_types, only: DP,    &  ! 倍精度実数型. Double precision.
    &                 TOKEN, &
    &                 STRING

  ! メッセージ出力
  ! Messgae dump
  !
  use dc_message, only: MessageNotify

  ! 数学
  ! Mathematic
  !
  use igmcore_math, only: PI

  ! 正二十面体格子データの管理
  ! Manager of Icosahedral grid data
  !
  use IcGrid2D_FVM_Manager, only: &
    & IcGrid2D_FVM, &
    & get_idMin, get_idMax, get_EffSize_Min, get_EffSize_Max, get_IcRadius, &
    & RC_REGIONS_NUM

  ! 物理場データを管理する構造体のための基底モジュール
  ! Base module for some derived types to manage the physical field data
  !
  use Physical_Field_Manager, only: &
    & Physical_Field, &
    & Physical_Field_Init, &
    & get_name_ => get_field_name, &
    & get_long_name_ => get_long_field_name, &
    & get_dimension_ => get_field_rank, &
    & get_units_ => get_field_units, &
    & set_name_ => set_field_name, &
    & set_long_name_ => set_long_field_name, &
    & set_units_ => set_field_units

  ! 宣言文 ; Declaration statement
  !
  implicit none
  private

  !
  !> \~japanese
  !! \brief 水平 2 次元的に分布する正二十面体格子上の物理場データを管理する構造体.
  !!
  !!  本クラスは, 正二十面体格子上の物理場データ, 物理場の名前, 単位などを管理する.
  !! この物理場データと対応付けられる正二十面体格子座標データは, 構造体 IcGrid2D_FVM が持っている.
  !! 本構造体は, その成分として, この構造体 IcGrid2D_FVM の変数のポインタを保持している.
  !!
  !! \~english
  !! \brief The derived type which manages the data of a physical field on a icosahedral grid distributed horizontally.
  !!
  !!  This derived type manages the physical field data on icosahedral grid and the name and unit of the field.
  !! The vertex data of icosahedral grid coressponded to this physical field data is contained in derived type IcGrid2D_FVM.
  !! This type contains the pointer to a variable of derived type IcGrid2D_FVM as a component of derived type IcGrid2D_FVM.
  !!
  type, public :: Field_IcGrid2D

    !> \~japanese
    !! \brief 正二十面体格子上で水平 2 次元的に分布した物理場データを保持する配列.
    !!
    !! 配列添字の各次元は, 正二十面体格子管理クラス IcGrid2D_FVM のメンバー変数である配列 rcs_AGrid の各次元と対応している.
    !! すなわち, それぞれの次元は,
    !! 矩形領域の ID, 水平方向の添え字, 水平方向の添え字, 物理場の各次元の成分, を意味する.
    !!
    !! \~english
    !! \brief 正二十面体格子上で水平 2 次元的に分布した物理場データを保持する配列.
    !!
    !! 配列添字の各次元は, 正二十面体格子管理クラス IcGrid2D_FVM のメンバー変数である配列 rcs_AGrid の各次元と対応している.
    !! すなわち, それぞれの次元は,
    !! 矩形領域の ID, 水平方向の添え字, 水平方向の添え字, 物理場の各次元の成分, を意味する.
    !!
    real(DP), pointer :: val(:,:,:,:) => null()
    ! f2003: real(DP), allocatable :: val(:,:,:,:)

    ! 非公開メンバ変数
    ! Private memeber variables
    !

    !> \~japanese 構造型 Field_IcGrid2D の基底となる構造体 Physical_Field の変数.
    !! \~english A variable of derived type Physcal_Field which is a base type of Field_IcGrid2D.
    type(Physical_Field) ::  base
    ! f2003: type(Physical_Field), private ::  base

    !> \~japanese 正二十面体格子を管理する構造体 IcGrid2D_FVM の変数へのポインタ.
    !! \~english The pointer to a variable of derived type IcGrid2D_FVM which manages the icosahedral grid.
    type(IcGrid2D_FVM), pointer :: icgrid
    ! f2003: type(IcGrid2D_FVM), pointer, private :: icgrid

  end type Field_IcGrid2D

  ! interface 文 ; interface statements
  !
  interface Field_IcGrid2D_Init
    module procedure init_Field_IcGrid2D1
    module procedure init_Field_IcGrid2D2
  end interface Field_IcGrid2D_Init

  ! 公開手続き
  ! Public procedures
  !

  public :: Field_IcGrid2D_Init, Field_IcGrid2D_Final
  public :: paste_field2D_margin, global_mean
  public :: get_icgrid, get_Physical_Field

  public :: get_field2D_name, get_long_field2D_name, get_field2D_rank, get_field2D_units
  public :: set_field2D_name, set_long_field2D_name, set_field2D_units

contains

!
!> \~japanese
!! \brief 構造型 Field_IcGrid2D の変数の初期化を行う.
!!
!! @param[in,out] self       構造体 Field_IcGrid2D の変数.
!! @param[in]     icgrid     構造体 IcGrid2D_FVM の変数.
!! @param[in]     field_rank  物理場のランク.
!! @param[in]     long_name  物理場の長い名前(あるいは物理場の解説文).
!! @param[in]     units      物理場の単位.
!!
!! \~english
!! \brief Initialize a variable of derived type Field_IcGrid2D.
!!
!! @param[in,out] self       The variable of derived type Field_IcGrid2D.
!! @param[in]     icgrid     The variable of derived type IcGrid2D_FVM.
!! @param[in]     name       The (short) name of the physical field.
!! @param[in]     field_rank  The rank of the physical filed.
!! @param[in]     long_name  The long name of (or the explanation for) the physical field.
!! @param[in]     units      The unit of the physical field.
!!
subroutine init_Field_IcGrid2D0( &
  & self,                     &  ! (inout)
  & icgrid, name, field_rank, &  ! (in)
  & long_name, units          &  ! (in)
  & )

  ! 宣言文 ; Declaration statement
  !
  type(Field_IcGrid2D), intent(inout) :: self
  type(IcGrid2D_FVM), pointer :: icgrid
  character(*), intent(in) :: name
  integer, intent(in) :: field_rank
  character(*), intent(in) :: long_name
  character(*), intent(in) :: units

  ! 作業変数
  ! Work variables
  !
  integer :: idMin, idMax

  ! 実行文 ; Executable statement
  !
!  write(*,*) 'Field_IcGrid2D initialize, name=', trim(name)

  self%icgrid => icgrid

  idMin = get_IdMin(self%icgrid)
  idMax = get_IdMax(self%icgrid)

  ! 物理場データを格納する配列のメモリを確保する.
  allocate(self%val(RC_REGIONS_NUM, idMin:idMax, idMin:idMax, 1:field_rank))

  ! 基底構造体の初期化関数を呼び出す.
  call Physical_Field_Init(self%base, name, field_rank, long_name, units)

end subroutine init_Field_IcGrid2D0

!
!> \~japanese
!! \brief 構造型 Field_IcGrid2D の変数の初期化を行う.
!!
!! @param[in,out] self       構造体 Field_IcGrid2D の変数.
!! @param[in]     icgrid     構造体 IcGrid2D_FVM の変数.
!! @param[in]     field_rank  物理場のランク.
!! @param[in]     long_name  物理場の長い名前(あるいは物理場の解説文).
!! @param[in]     units      物理場の単位.
!!
!! \~english
!! \brief Initialize a variable of derived type Field_IcGrid2D.
!!
!! @param[in,out] self       The variable of derived type Field_IcGrid2D.
!! @param[in]     icgrid     The variable of derived type IcGrid2D_FVM.
!! @param[in]     name       The (short) name of the physical field.
!! @param[in]     field_rank  The rank of the physical filed.
!! @param[in]     long_name  The long name of (or the explanation for) the physical field.
!! @param[in]     units      The unit of the physical field.
!!
subroutine init_Field_IcGrid2D1( &
  & self,                                &  ! (inout)
  & icgrid, name, rank, long_name, units &  ! (in)
  & )

  ! 宣言文 ; Declaration statement
  !
  type(Field_IcGrid2D), intent(inout) :: self
  type(IcGrid2D_FVM), intent(in), target :: icgrid
  character(*), intent(in) :: name
  integer, intent(in) :: rank
  character(*), intent(in) :: long_name
  character(*), intent(in) :: units

  ! 作業変数
  ! Work variables
  !
  integer :: idMin, idMax
  type(IcGrid2D_FVM), pointer :: icgrid_ptr

  ! 実行文 ; Executable statement
  !

  !  write(*,*) 'Field_IcGrid2D initialize, name=', trim(name)

  !
  icgrid_ptr => icgrid
  call init_Field_IcGrid2D0(self, icgrid_ptr, name, rank, long_name, units)

end subroutine init_Field_IcGrid2D1

!
!> \~japanese
!! \brief 構造型 Field_IcGrid2D の変数の初期化を行う.
!!
!! @param[in,out] self       構造体 Field_IcGrid2D の変数.
!! @param[in]     icgrid     構造体 IcGrid2D_FVM の変数.
!! @param[in]     long_name  物理場の長い名前(あるいは物理場の解説文).
!! @param[in]     units      物理場の単位.
!!
!! \~english
!! \brief Initialize a variable of derived type Field_IcGrid2D.
!!
!! @param[in,out] self       The variable of derived type Field_IcGrid2D.
!! @param[in]     icgrid     The variable of derived type IcGrid2D_FVM.
!! @param[in]     name       The (short) name of the physical field.
!! @param[in]     field_rank  The rank of the physical filed.
!!
subroutine init_Field_IcGrid2D2( &
  & self,              &  ! (inout)
  & icgrid, name, rank &  ! (in)
  & )

  ! 宣言文 ; Declaration statement
  !
  type(Field_IcGrid2D), intent(inout) :: self
  type(IcGrid2D_FVM), intent(in), target :: icgrid
  character(*), intent(in) :: name
  integer, intent(in) :: rank

  ! 作業変数
  ! Work variables
  !

  ! 実行文 ; Executable statement
  !

  call init_Field_IcGrid2D1(self, icgrid, name, rank, '', '')

end subroutine init_Field_IcGrid2D2

!
!> \~japanese
!! \brief 構造型 Field_IcGrid2D の変数を最終化する.
!!
!! @param[in,out] self 構造型 Field_IcGrid2D の変数.
!!
!! \~english
!! \brief Finalizes a variable of derived type Field_IcGrid2D.
!!
!! @param[in,out] self A variable of derived type Field_IcGrid2D.
!!
subroutine Field_IcGrid2D_Final( &
  & self &  ! (inout)
  & )

  ! 宣言文 ; Declaration statement
  !
  type(Field_IcGrid2D), intent(inout) :: self

  ! 実行文　; Executable statement
  !

!  write(*,*) 'Field_IcGrid2D finalize, name=', trim(get_name(self))

  ! 物理場データを保持している配列のメモリを解放する.
  deallocate(self%val)

end subroutine Field_IcGrid2D_Final

!
!> \~japanese
!! \brief 構造型 Field_IcGrid2D の基底構造型 Physical_Field の変数へのポインタを取得する.
!!
!! @param[in,out] self 構造型 Field_IcGrid2D の変数.
!! @return 基底構造型 Physical_Field の変数へのポインタ.
!!
!! \~english
!! \brief Gets the pointer to a variable of derived type Physical_Field which is a base derived type of Field_IcGrid2D.
!!
!! @param[in,out] self A variable of derived type Field_IcGrid2D.
!! @return The pointer to a variable of base derived type  Physical_Field.
!!
function get_Physical_Field( &
  & self &  ! (inout)
  & ) result(val)

  ! 宣言文 ; Declaration statements
  !
  type(Field_IcGrid2D), intent(inout), target :: self
  type(Physical_Field), pointer :: val

  ! 実行文 ; Executable statements
  !
  val => self%base

end function get_Physical_Field

!
!> \~japanese
!! \brief 構造体 Field_IcGrid2D の成分である構造型 IcGrid2D_FVM の変数へのポインタを取得する.
!!
!! @param[in,out] self 構造型 Field_IcGrid2D の変数.
!! @return 構造型 IcGrid2D_FVM の変数へのポインタ.
!!
!! \~english
!! \brief Get the pointer to the variable of derived type IcGrid2D_FVM which is a component of derived type Field_IcGrid2D.
!!
!! @param[in,out] self Field_IcGrid2D A variable of derived type Field_IcGrid2D.
!! @return IcGrid2D_FVM The pointer to a variable of derived type IcGrid2D_FVM.
!!
function get_icgrid( &
  & self &  ! (inout)
  & ) result(icgrid)

  ! 宣言文 ; Declaration statement
  !
  type(Field_IcGrid2D), intent(in) :: self       !
  type(IcGrid2D_FVM), pointer :: icgrid

  ! 実行文 ; Executable statement
  !

  icgrid => self%icgrid

end function get_icgrid

!
!> \~japanese
!! \brief 全矩形領域の物理場データを保持する配列に対して, 袖領域を隣り合う矩形領域の境界近傍の物理データで埋める処理を行う.
!!
!! @param[in,out] self 構造型 Field_IcGrid2D の変数.
!!
!! \~english
!! \brief 全矩形領域の物理場データを保持する配列に対して, 袖領域を隣り合う矩形領域の境界近傍の物理データで埋める処理を行う.
!!
!! @param[in,out] self Field_IcGrid2D A variable of derived type Field_IcGrid2D.
!!
subroutine paste_field2D_margin( &
  & self &  ! (inout)
  & )

  ! 宣言文 ; Declaration statements
  !
  type(Field_IcGrid2D), intent(inout) :: self       !

  ! 作業変数
  ! Work variables
  !
  integer :: idlist1(7) = (/ 5, 1, 2, 3, 4, 5, 1 /)
  integer :: idlist2(7) = (/ 10, 6, 7, 8, 9, 10, 6 /)
  integer :: EMin, EMax, marginw, idMin
  integer i, k

  ! 実行文 ; Executable statements
  !

  EMin = get_EffSize_Min(self%icgrid)
  EMax = get_EffSize_Max(self%icgrid)

  ! ノリシロの幅を計算する.
  idMin = get_IdMin(self%icgrid)
  marginw = EMin - idMin

  do i=1, RC_REGIONS_NUM/2
    do k=1, marginw
      ! 北半球側の矩形領域ののりしろに値をコピーする.
      self%val(i, EMin:EMax, EMin-k, :) = self%val(idlist1(i), EMin+k, EMin:EMax, :)
      self%val(i, EMax+k, EMin:EMax, :) = self%val(idlist2(i), EMin+k, EMin:EMax, :)
      self%val(i, EMin:EMax, EMax+k, :) = self%val(idlist2(i+1), EMin:EMax, EMin+k, :)
      self%val(i, EMin-k, EMin:EMax, :) = self%val(idlist1(i+2), EMin:EMax, EMin+k, :)

      ! 南半球側の矩形領域ののりしろに値をコピーする.
      self%val(i+5, EMin:EMax, EMin-k, :) = self%val(idlist1(i+1), EMin:EMax, EMax-k, :)
      self%val(i+5, EMax+k, EMin:EMax, :) = self%val(idlist2(i), EMin:EMax, EMax-k, :)
      self%val(i+5, EMin:EMax, EMax+k, :) = self%val(idlist2(i+2), EMax-k, EMin:EMax, :)
      self%val(i+5, EMin-k, EMin:EMax, :) = self%val(idlist1(i+2), EMax-k, EMin:EMax, :)
    end do
  end do

end subroutine paste_field2D_margin

!
!> \~japanese
!! \brief 正二十面体格子点上の物理場の全球空間平均を計算する.
!!
!! @param[in,out] self 構造型 Field_IcGrid2D の変数.
!! @return 物理場の全球空間平均値.
!!
!! \~english
!! \brief Calculate the global spatial average of a physical field on icosahedral grid.
!!
!! @param[in,out] self A variable of derived type Field_IcGrid2D.
!! @return The global spatial average of a physical field.
!!
function global_mean( &
  & self &  ! (inout)
  & ) result(val)

  ! 宣言文 ; Declaration statement
  !
  type(Field_IcGrid2D), intent(inout) :: self       !
  real(DP)  val

  ! 作業変数
  ! Work variables
  !
  integer :: rcID, i, j
  integer :: EMin, EMax
  type(IcGrid2D_FVM), pointer :: icgrid

  ! 実行文 ; Executable statement
  !

  ! 物理場のランクが 1 次元(スカラー場)であることを確認する.
  if( get_field2D_rank(self) /= 1 ) then
    call MessageNotify( 'E', 'IGMBaseLib IO::Field_IcGrid2D::global_mean', &
      & "The rank of Field_IcGrid2D(=%d) is invalid.", i=(/ get_field2D_rank(self) /) )
    stop
    return
  end if

  EMin = get_EffSize_Min(self%icgrid)
  EMax = get_EffSize_Max(self%icgrid)
  icgrid => self%icgrid

  ! まず指定されたスカラー場の, 全球積分値を求める.
  val = self%val(1,EMin,EMin, 1) * icgrid%rcs_CVArea(1,EMin,EMin) &
        + self%val(10,EMax,EMax, 1)  * icgrid%rcs_CVArea(10,EMax,EMax)

  do rcID=1, RC_REGIONS_NUM
    do j=EMin, EMax-1
      do i=EMin+1, EMax
        val = val + self%val(rcID,i,j, 1) * icgrid%rcs_CVArea(rcID,i,j)
      end do
    end do
  end do

  ! 球の面積で規格化する.
  val = val / ( 4.0d0 * PI * get_IcRadius(icgrid)**2 )

end function global_mean

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! Physical_Field_mod の手続きをオーバライド
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


!
!> \~japanese
!! \brief 物理場の長い名前(あるいは解説文)を取得する.
!!
!! @param[in] self 構造型 Field_IcGrid2D の変数.
!! @return  物理場の長い名前(あるいは解説文).
!!
!! \~english
!! \brief Get the long name of (or explanation for) the physcial field.
!!
!! @param[in] self The variable of derived type Field_IcGrid2D.
!! @return  The long name of (or explanation for) the physical field.
!!
function get_long_field2D_name(self) result(val)

  ! 宣言文 ; Declaration statement
  !
  type(Field_IcGrid2D), intent(in) :: self
  character(STRING) val

  ! 実行文 ; Executable statement
  !

  val = get_long_name_(self%base)

end function get_long_field2D_name

!
!> \~japanese
!! \brief 物理場の長い名前(あるいは解説文)を設定する.
!!
!! @param[in,out] self 構造型 Field_IcGrid2D の変数.
!! @param[in] long_name 物理場の長い名前(あるいは解説文).
!!
!! \~english
!! \brief Set the long name of (or explanation for) the physcial field.
!!
!! @param[in,out] self The variable of derived type Field_IcGrid2D.
!! @param[in] long_name The long name of (or explanation for) the physical field.
!!
subroutine set_long_field2D_name(self, long_name)
  ! 宣言文 ; Declaration statement
  !
  type(Field_IcGrid2D), intent(inout) :: self
  character(*), intent(in) ::  long_name

  ! 実行文 ; Executable statement
  !

  call set_long_name_(self%base, long_name)

end subroutine set_long_field2D_name

!
!> \~japanese
!! \brief 物理場の単位を取得する.
!!
!! @param[in] self 構造型 Field_IcGrid2D の変数.
!! @reeturn    物理場の単位.
!!
!! \~english
!! \brief Get the unit of the physical field.
!!
!! @param[in] self The variable of derived type Field_IcGrid2D.
!! @return    The unit of the physical field.
!!
function get_field2D_units(self) result(val)
  ! 宣言文 ; Declaration statement
  !
  type(Field_IcGrid2D), intent(in) :: self
  character(TOKEN) val

  ! 実行文 ; Executable statement
  !

  val = get_units_(self%base)

end function get_field2D_units

!
!> \~japanese
!! \brief 物理場の単位を設定する.
!!
!! @param[in,out] self 構造型 Field_IcGrid2D の変数.
!! @param[in] units 物理場の単位.
!!
!! \~english
!! \brief Sets the unit of the physical field.
!!
!! @param[in,out] self The variable of derived type Field_IcGrid2D.
!! @param[in] units The unit of the physical field.
!!
subroutine set_field2D_units(self, units)
  ! 宣言文 ; Declaration statement
  !
  type(Field_IcGrid2D), intent(inout) :: self
  character(*), intent(in) ::  units

  ! 実行文 ; Executable statement
  !

  call set_units_(self%base, units)

end subroutine set_field2D_units

!
!> \~japanese
!! \brief 物理場の次元を取得する.
!!
!! @param[in] self 構造型 Field_IcGrid2D の変数.
!! @reeturn    物理場のランク.
!!
!! \~english
!! \brief Get The rank of the physical field.
!!
!! @param[in] self The variable of derived type Field_IcGrid2D.
!! @reeturn    The rank of the physical field.
!!
function get_field2D_rank(self) result(val)
  ! 宣言文 ; Declaration statement
  !
  type(Field_IcGrid2D), intent(in) :: self
  integer val

  ! 実行文 ; Executable statement
  !

  val = get_dimension_(self%base)

end function get_field2D_rank


!
!> \~japanese
!! \brief 物理場の(短い)名前を取得する.
!!
!! @param[in] self 構造型 Field_IcGrid2D の変数.
!! @return 物理場の(短い)名前.
!!
!! \~english
!! \brief Get the (short) name of the physical field.
!!
!! @param[in] self The variable of derived type Field_IcGrid2D.
!! @return The (short) name of the physical field.
!!
function get_field2D_name( &
  & self &
  & ) result(name)

  ! 宣言文 ; Declaration statement
  !
  type(Field_IcGrid2D), intent(in) :: self       !
  character(TOKEN) name

  ! 実行文 ; Executable statement
  !

  name = get_name_(self%base)

end function get_field2D_name

!
!> \~japanese
!! \brief 物理場の(短い)名前を設定する.
!!
!! @param[in,out] self 構造型 Field_IcGrid2D の変数.
!! @param[in] name 物理場の(短い)名前.
!!
!! \~english
!! \brief Sets the (short) name of the physical field.
!!
!! @param[in,out] self The variable of derived type Field_IcGrid2D.
!! @param[in] name The (short) name of the physical field.
!!
subroutine set_field2D_name( &
  & self, name )

  ! 宣言文 ; Declaration statement
  !
  type(Field_IcGrid2D), intent(inout) :: self
  character(*), intent(in) ::  name

  ! 実行文 ; Executable statement
  !

  call set_name_(self%base, name)

end subroutine set_field2D_name

end module Field_IcGrid2D_Manager
